/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mouse Gestures for Mozilla.
 *
 * The Initial Developer of the Original Code is Jens Bannmann.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s) (alphabetical order):
 *  Jens Bannmann <jens.b@web.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const RDFService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
                             .getService(Components.interfaces.nsIRDFService);

const RDFCUtils = Components.classes["@mozilla.org/rdf/container-utils;1"]
                  .getService().QueryInterface(Components.interfaces.nsIRDFContainerUtils);

const mgNs = "http://optimoz.mozdev.org/gestures-rdf#";

var mgUserDS;
var mgDefaultMappingsDS;

var mgUpdateTimer;

function getUserRDF() {
  if (mgUserDS) {
    return mgUserDS;
  } else {
    var mgUserFile = getProfileDir();
    mgUserFile.append("mousegestures.rdf");
    mgUserDS = RDFService.GetDataSourceBlocking(fileToURL(mgUserFile).spec);
    return mgUserDS;
  }
}

function getDefaultMappings() {
  if (!mgDefaultMappingsDS) {
    mgDefaultMappingsDS = RDFService
        .GetDataSourceBlocking("chrome://mozgest/content/defaultMappings.rdf");
  }
  return mgDefaultMappingsDS;
}

function initUserRDF() {
  var dsUser = getUserRDF();

  var usInfo = RDFService.GetResource("urn:mozgest:userstore");
  // determine which mozgest-version saved this userstore. default to 0 if 
  //    userstore doesnt exist - this will trigger installation of mappings below
  var userstoreVersion = getPropertyValue(dsUser, usInfo, mgNs + "creatorVersion") || 0;

  var dsDefault = getDefaultMappings();
  var releaseInfo = RDFService.GetResource("urn:mozgest:release");
  var releaseVersion = getPropertyValue(dsDefault, releaseInfo, mgNs + "version");

  if (releaseVersion > userstoreVersion) {
    unassertLiteral(dsUser, usInfo, mgNs + "creatorVersion", userstoreVersion);
    assertLiteral(dsUser, usInfo, mgNs + "creatorVersion", releaseVersion);
    // enumerate all resources from defaultMappings file
    var elements = dsDefault.GetAllResources();
    while (elements.hasMoreElements()) {
      var el = elements.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
      // only process Seqs
      if(RDFCUtils.IsSeq(dsDefault, el)) {
//        var seqRes = RDFService.GetResource(el.Value);
        var seqRes = el;
        var sourceSeq = newContainer();
        sourceSeq.Init(dsDefault, el);
        var targetSeq = newContainer();
        // init current seq in userstore, or create it if it doesn't exist
        var emptySeq = false;
        try {
          targetSeq.Init(dsUser, seqRes);
        } catch (e) {
          RDFCUtils.MakeSeq(dsUser, seqRes);
          targetSeq.Init(dsUser, seqRes);
          emptySeq = true;
        }
        var mappingsEnum = sourceSeq.GetElements();
        while (mappingsEnum.hasMoreElements()) {
          // get mapping data
          var mapping = mappingsEnum.getNext();

          // determine which mozgest version introduced this mapping.
          //    default to 0.4 if no "since" property present
          var mSince = getPropertyValue(dsDefault, mapping, mgNs + "since") || 0.4;

          // if mapping is newer than userstoreVersion, install it
          if (mSince > userstoreVersion) {
            if (emptySeq ||
                !findMapping(getPropertyValue(dsDefault,
                                              mapping,
                                              mgNs + "code"), targetSeq)) {
              installMapping(mapping, targetSeq);
            }
          }
        }
      }
    }
    dsUser.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
  }
  dsUser.AddObserver(mgMappingsObserver);
}

// lookup mapping by code
function findMapping(code, targetContainer) {
  var dsUser = getUserRDF();
  var mappingsEnum = targetContainer.GetElements();
  while (mappingsEnum.hasMoreElements()) {
    var mapping = mappingsEnum.getNext();
    if (getPropertyValue(dsUser, mapping, mgNs + "code") == code) {
      return mapping;
    }
  }
  return false;
}

// installs a mapping from defaultMappings to userstore.
function installMapping(mapping, targetContainer) {
  var dsUser = getUserRDF();
  var dsDefault = getDefaultMappings();

  var mCode = getPropertyValue(dsDefault, mapping, mgNs + "code");
  var mFunction = getPropertyValue(dsDefault, mapping, mgNs + "function");
  var mName = bundle.getString(mFunction);
  var mAppearance = getPropertyValue(dsDefault, mapping, mgNs + "appearance");

  node = RDFService.GetAnonymousResource();
  targetContainer.AppendElement(node, true);
  assertLiteral(dsUser, node, mgNs + "code", mCode);
  assertLiteral(dsUser, node, mgNs + "name", mName);
  assertLiteral(dsUser, node, mgNs + "function", mFunction);
  if (mAppearance) {
    assertLiteral(dsUser, node, mgNs + "appearance", mAppearance);
  }
}

function readRDFGestureTable(component) {
  // get our datasource
  var dsource = getUserRDF();

  // instantiate a container object and its root node
  var mappingsContainer = newContainer();
  var rootnode = RDFService.GetResource("urn:mozgest:mappings:" + component);

  try {
    mappingsContainer.Init(dsource, rootnode);
  } catch (e) {
    // no existing mappings
    return;
  }

  var enumerator = mappingsContainer.GetElements();
  while (enumerator.hasMoreElements()) {
    var mapping = enumerator.getNext()
                  .QueryInterface(Components.interfaces.nsIRDFResource);
    var entry = new Mapping();
    entry.resource = mapping.Value;
    entry.name = getPropertyValue(dsource, mapping, mgNs + "name");
    entry.code = getPropertyValue(dsource, mapping, mgNs + "code");
    entry.appearance = getPropertyValue(dsource, mapping, mgNs + "appearance");
    entry.func = getPropertyValue(dsource, mapping, mgNs + "function");
    entry.custom = getPropertyValue(dsource, mapping, mgNs + "custom");
    entry.countBase = Math.abs(getPropertyValue(dsource, mapping, mgNs + "count")) || 0;
    addGesture(entry);
  }
}


function saveGestureCounts(e) {
  // do not react on html page-loads (they fire "unloads" as well)
  if (e.view == window) {
    dump("MozGest: Saving gesture counts\n");
    var dsource = getUserRDF();
    for (var code in gesturesTable) {
      mapping = gesturesTable[code];
      if (mapping.countCur != 0) {
        // get currently saved value
        var node = RDFService.GetResource(mapping.resource);
        var old = Math.abs(getPropertyValue(dsource, node, mgNs + "count")) || 0;
  
        // add current session count and save it
        var n = old + mapping.countCur;
        changeLiteral(dsource, node, mgNs + "count", old, n);
      }
    }
  }
}


/*** convenience functions for RDF access ***/

function newContainer() {
  return Components.classes['@mozilla.org/rdf/container;1'].createInstance()
         .QueryInterface(Components.interfaces.nsIRDFContainer);
}

function assertLiteral(dsource, node, propName, value) {
  dsource.Assert(node, RDFService.GetResource(propName), 
                 RDFService.GetLiteral(value), true);
}

function unassertLiteral(dsource, node, propName, value) {
  dsource.Unassert(node, RDFService.GetResource(propName), 
                 RDFService.GetLiteral(value), true);
}

function changeLiteral(dsource, node, propName, oVal, nVal) {
  dsource.Change(node, RDFService.GetResource(propName), 
                 RDFService.GetLiteral(oVal), RDFService.GetLiteral(nVal));
}

function getPropertyValue(dsource, obj, propName) {
  // get nsIRDFNode
  var literal = dsource.GetTarget(obj, RDFService.GetResource(propName), true);
  if (literal) {
    // convert to nsIRDFLiteral and get the value
    return literal.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
  } else {
    return null;
  }
}


/*** profile & URL handling ***/

function getProfileDir() {
  // get a File representing user's profile dir
  var profileInternal = Components.classes["@mozilla.org/profile/manager;1"]
      .createInstance().QueryInterface(Components.interfaces.nsIProfileInternal);
  return profileInternal.getProfileDir(profileInternal.currentProfile);
}


function fileToURL(aFile) {
  var fileHandler = Components.classes["@mozilla.org/network/io-service;1"]
                      .getService(Components.interfaces.nsIIOService);
  return fileHandler.newFileURI(aFile);
}


/*** observer stuff ***/

var mgMappingsObserver = new Object();
mgMappingsObserver = {
  onAssert: function(aDataSource, aSource, aProperty, aTarget) {
    if (aProperty.Value.indexOf(mgNs) == 0) {
      delayedUpdate();
    }
  },

  onUnassert : function(aDataSource, aSource, aProperty, aTarget) {
    if (aProperty.Value.indexOf(mgNs) == 0) {
      delayedUpdate();
    }
  },

  onChange: function(aDataSource, aSource, aProperty, aTarget, aOldTarget,
                     aNewTarget) {
    if (aProperty.Value.indexOf(mgNs) == 0
        && aProperty.Value != mgNs + "count") {
      delayedUpdate();
    }
  },

  onMove: function(aDataSource, aOldSource, aNewSource, aProperty,
  aTarget) {
  },

  beginUpdateBatch: function(aDataSource) {
  },

  endUpdateBatch: function(aDataSource) {
  }
}

function delayedUpdate() {
  window.clearTimeout(mgUpdateTimer);
  mgUpdateTimer = window.setTimeout(loadGestureTable, 500);
}
